@@ -2,10 +2,10 @@
   (:require [puppetlabs.puppetserver.cli.subcommand :as cli]
             [puppetlabs.services.jruby.jruby-puppet-core :as jruby-core]))
 
-(defn run!
+(defn gem-run!
   [config args]
   (jruby-core/cli-run! config "gem" args))
 
 (defn -main
   [& args]
-  (cli/run run! args))
+  (cli/run gem-run! args))
