@@ -1,21 +1,20 @@
 (ns puppetlabs.trapperkeeper.services.webserver.jetty9-core
   (:import (org.eclipse.jetty.server Handler Server Request ServerConnector
                                      HttpConfiguration HttpConnectionFactory
-                                     ConnectionFactory)
+                                     ConnectionFactory AbstractConnectionFactory)
            (org.eclipse.jetty.server.handler AbstractHandler ContextHandler HandlerCollection
                                              ContextHandlerCollection AllowSymLinkAliasChecker StatisticsHandler)
            (org.eclipse.jetty.util.resource Resource)
            (org.eclipse.jetty.util.thread QueuedThreadPool)
            (org.eclipse.jetty.util.ssl SslContextFactory)
-           (javax.servlet.http HttpServletRequest HttpServletResponse)
-           (java.util.concurrent Executors TimeoutException)
+           (javax.servlet.http HttpServletResponse)
+           (java.util.concurrent TimeoutException)
            (org.eclipse.jetty.servlets.gzip GzipHandler)
            (org.eclipse.jetty.servlet ServletContextHandler ServletHolder DefaultServlet)
            (org.eclipse.jetty.webapp WebAppContext)
            (java.util HashSet)
-           (org.eclipse.jetty.http MimeTypes)
+           (org.eclipse.jetty.http MimeTypes HttpHeader HttpHeaderValue)
            (javax.servlet Servlet ServletContextListener)
-           (java.io File IOException)
            (org.eclipse.jetty.proxy ProxyServlet)
            (java.net URI)
            (java.security Security)
@@ -23,15 +22,15 @@
            (clojure.lang Atom)
            (java.lang.management ManagementFactory)
            (org.eclipse.jetty.jmx MBeanContainer)
-           (org.eclipse.jetty.util URIUtil))
+           (org.eclipse.jetty.util URIUtil BlockingArrayQueue)
+           (java.io IOException))
 
   (:require [ring.util.servlet :as servlet]
             [ring.util.codec :as codec]
             [clojure.string :as str]
-            [clojure.set :as set]
             [clojure.tools.logging :as log]
-            [puppetlabs.kitchensink.core :as ks]
             [puppetlabs.trapperkeeper.services.webserver.jetty9-config :as config]
+            [puppetlabs.trapperkeeper.services.webserver.experimental.jetty9-websockets :as websockets]
             [schema.core :as schema]))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -88,12 +87,9 @@
     (schema/optional-key :callback-fn) (schema/pred ifn?)
     (schema/optional-key :failure-callback-fn) (schema/pred ifn?)
     (schema/optional-key :request-buffer-size) schema/Int
-    (schema/optional-key :follow-redirects) schema/Bool))
-
-(def ServerContext
-  {:state     Atom
-   :handlers  ContextHandlerCollection
-   :server    (schema/maybe Server)})
+    (schema/optional-key :follow-redirects) schema/Bool
+    (schema/optional-key :idle-timeout) (schema/both schema/Int
+                                                     (schema/pred pos?))))
 
 (def ContextEndpoint
   {:type                                    (schema/eq :context)
@@ -103,6 +99,9 @@
 (def RingEndpoint
   {:type     (schema/eq :ring)})
 
+(def WebsocketEndpoint
+  {:type     (schema/eq :websocket)})
+
 (def ServletEndpoint
   {:type     (schema/eq :servlet)
    :servlet  java.lang.Class})
@@ -121,6 +120,7 @@
   (schema/conditional
     #(-> % :type (= :context)) ContextEndpoint
     #(-> % :type (= :ring)) RingEndpoint
+    #(-> % :type (= :websocket)) WebsocketEndpoint
     #(-> % :type (= :servlet)) ServletEndpoint
     #(-> % :type (= :war)) WarEndpoint
     #(-> % :type (= :proxy)) ProxyEndpoint))
@@ -128,10 +128,17 @@
 (def RegisteredEndpoints
   {schema/Str [Endpoint]})
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
-;;; Constants
+(def ServerContextState
+  {:mbean-container (schema/maybe MBeanContainer)
+   :overrides-read-by-webserver schema/Bool
+   :overrides (schema/maybe {schema/Keyword schema/Any})
+   :endpoints RegisteredEndpoints
+   :ssl-context-factory (schema/maybe SslContextFactory)})
 
-(def default-graceful-stop-timeout 60000)
+(def ServerContext
+  {:state     (schema/atom ServerContextState)
+   :handlers  ContextHandlerCollection
+   :server    (schema/maybe Server)})
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Utility Functions
@@ -183,7 +190,15 @@
                   (.setKeyStore (:keystore keystore-config))
                   (.setKeyStorePassword (:key-password keystore-config))
                   (.setTrustStore (:truststore keystore-config))
+                  ;; Need to clear out the default cipher suite exclude list so
+                  ;; that Jetty doesn't potentially remove one or more ciphers
+                  ;; that we want to be included.
+                  (.setExcludeCipherSuites (into-array String []))
                   (.setIncludeCipherSuites (into-array String cipher-suites))
+                  ;; Need to clear out the default protocols exclude list so
+                  ;; that Jetty doesn't potentially remove one or more protocols
+                  ;; that we want to be included.
+                  (.setExcludeProtocols (into-array String []))
                   (.setIncludeProtocols (into-array String protocols)))]
     (if (:trust-password keystore-config)
       (.setTrustStorePassword context (:trust-password keystore-config)))
@@ -212,13 +227,48 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; Jetty Server / Connector Functions
 
-(defn- connection-factory
+(defn- http-configuration
   [request-header-size]
   (let [http-config (doto (HttpConfiguration.)
-                      (.setSendDateHeader true)
-                      (.setRequestHeaderSize request-header-size))]
-    (into-array ConnectionFactory
-                [(HttpConnectionFactory. http-config)])))
+                      (.setSendDateHeader true))]
+    (if request-header-size
+      (.setRequestHeaderSize http-config request-header-size))
+    http-config))
+
+(defn- connection-factories
+  [request-header-size ssl-ctxt-factory]
+  (let [http-config (http-configuration request-header-size)
+        factories   (into-array ConnectionFactory
+                                [(HttpConnectionFactory. http-config)])]
+    (if ssl-ctxt-factory
+      (AbstractConnectionFactory/getFactories
+        ssl-ctxt-factory factories)
+      factories)))
+
+(defn- thread-count
+  [setting]
+  (if setting setting -1))
+
+(schema/defn ^:always-validate
+  connector* :- ServerConnector
+  [server :- Server
+   config :- (merge config/WebserverConnector
+                    {schema/Keyword schema/Any})
+   ssl-ctxt-factory :- (schema/maybe SslContextFactory)]
+  (let [request-size (:request-header-max-size config)
+        connector   (doto (ServerConnector.
+                            server
+                            nil nil nil
+                            (thread-count (:acceptor-threads config))
+                            (thread-count (:selector-threads config))
+                            (connection-factories request-size ssl-ctxt-factory))
+                      (.setPort (:port config))
+                      (.setHost (:host config)))]
+    (when-let [linger-millis (:so-linger-milliseconds config)]
+      (.setSoLingerTime connector linger-millis))
+    (when-let [idle-timeout (:idle-timeout-milliseconds config)]
+      (.setIdleTimeout connector idle-timeout))
+    connector))
 
 (schema/defn ^:always-validate
   ssl-connector  :- ServerConnector
@@ -226,31 +276,76 @@
   [server            :- Server
    ssl-ctxt-factory  :- SslContextFactory
    config :- config/WebserverSslConnector]
-  (let [request-size (:request-header-max-size config)]
-    (doto (ServerConnector. server ssl-ctxt-factory (connection-factory request-size))
-      (.setPort (:port config))
-      (.setHost (:host config)))))
+  (connector* server config ssl-ctxt-factory))
 
 (schema/defn ^:always-validate
   plaintext-connector :- ServerConnector
   [server :- Server
    config :- config/WebserverConnector]
-  (let [request-size (:request-header-max-size config)]
-    (doto (ServerConnector. server (connection-factory request-size))
-      (.setPort (:port config))
-      (.setHost (:host config)))))
+  (connector* server config nil))
+
+(schema/defn ^:always-validate
+  queue-thread-pool :- (schema/maybe QueuedThreadPool)
+  [max-threads :- (schema/maybe schema/Int)
+   queue-max-size :- (schema/maybe schema/Int)]
+  (if (or max-threads queue-max-size)
+    (let [thread-pool (if max-threads
+                        (QueuedThreadPool. max-threads)
+                        (QueuedThreadPool.))]
+      (if queue-max-size
+        ;; The code below is definitely not ideal, but there isn't a way to set
+        ;; the maximum capacity of the QueuedThreadPool's BlockingArrayQueue
+        ;; after construction.  We're trying to avoid hard-coding our own
+        ;; defaults for other settings that we want Jetty to control, e.g., the
+        ;; initial capacity of the queue and minimum number of threads.  By
+        ;; reconstructing the QueuedThreadPool here, we can use Jetty's defaults
+        ;; for settings unrelated to `queue-max-size`.
+        ;;
+        ;; QueuedThreadPool and BlockingArrayQueue construction isn't too
+        ;; expensive.  It mostly involves some initial memory allocations.  The
+        ;; more expensive work - where threads are actually started and the
+        ;; queue expands to fulfill new requests - would only happen after the
+        ;; QueuedThreadPool were started.  That won't happen for the
+        ;; `thread-pool` instance from above, which just gets thrown away as
+        ;; this function falls out of scope without ever having been started.
+        ;; Also, this function would likely only be called once per server
+        ;; startup where a `queue-max-size` were configured.
+        ;;
+        ;; The QueuedThreadPool constructor sets the `queue-capacity` and
+        ;; `queue-grow-by` based on the minimum number of threads available
+        ;; in the pool.  See https://github.com/eclipse/jetty.project/blob/jetty-9.2.10.v20150310/jetty-util/src/main/java/org/eclipse/jetty/util/thread/QueuedThreadPool.java#L92-L96.
+        ;; That algorithm is essentially duplicated here, with the only
+        ;; difference being that if `queue-max-size` is smaller than the
+        ;; minimum number of threads, the `queue-capacity` and `queue-grow-by`
+        ;; are reduced to the `queue-max-size` in order to avoid the
+        ;; BlockingArrayQueue constructor throwing an IllegalArgumentException.
+        (let [min-threads    (.getMinThreads thread-pool)
+              queue-capacity (min queue-max-size min-threads)
+              queue-grow-by  (min queue-max-size min-threads)]
+          (QueuedThreadPool. (.getMaxThreads thread-pool)
+                             min-threads
+                             (.getIdleTimeout thread-pool)
+                             (BlockingArrayQueue.
+                               queue-capacity
+                               queue-grow-by
+                               queue-max-size)))
+        thread-pool))))
 
 (schema/defn ^:always-validate
   create-server :- Server
   "Construct a Jetty Server instance."
   [webserver-context :- ServerContext
    config :- config/WebserverConfig]
-  (let [server (Server. (QueuedThreadPool. (:max-threads config)))]
+  (let [server (if-let [pool (queue-thread-pool (:max-threads config)
+                                                (:queue-max-size config))]
+                 (Server. pool)
+                 (Server.))]
     (when (:jmx-enable config)
       (let [mb-container (MBeanContainer. (ManagementFactory/getPlatformMBeanServer))]
         (doto server
           (.addEventListener mb-container)
-          (.addBean mb-container))))
+          (.addBean mb-container))
+        (swap! (:state webserver-context) assoc :mbean-container mb-container)))
     (when (:http config)
       (let [connector (plaintext-connector server (:http config))]
         (.addConnector server connector)))
@@ -306,6 +401,15 @@
    enable-trailing-slash-redirect?]
   (.setAllowNullPathInfo handler (not enable-trailing-slash-redirect?))
   (.addHandler (:handlers webserver-context) handler)
+  ;; If this handler is being added after the server has been started, we
+  ;; need to mark the handler as "managed" so that the server will stop the
+  ;; handler when the server is stopped.  We also need to manually start the
+  ;; handler.  The server takes care of marking handlers as managed and starting
+  ;; them if the handlers are already registered when the server is started.
+  (if-let [server (.getServer handler)]
+    (when (and (.isRunning server) (not (.isRunning handler)))
+      (.manage (:handlers webserver-context) handler)
+      (.start handler)))
   handler)
 
 (defn- ring-handler
@@ -327,8 +431,9 @@
    target :- ProxyTarget
    options :- ProxyOptions]
   (let [custom-ssl-ctxt-factory (when (map? (:ssl-config options))
-                                  (get-proxy-client-context-factory (:ssl-config options)))
-        request-buffer-size     (:request-buffer-size options)]
+                                  (get-proxy-client-context-factory
+                                    (:ssl-config options)))
+        {:keys [request-buffer-size idle-timeout]} options]
     (proxy [ProxyServlet] []
       (rewriteURI [req]
         (let [query (.getQueryString req)
@@ -357,36 +462,57 @@
       (newHttpClient []
         (let [client (if custom-ssl-ctxt-factory
                        (HttpClient. custom-ssl-ctxt-factory)
-                       (if-let [ssl-ctxt-factory (:ssl-context-factory @(:state webserver-context))]
+                       (if-let [ssl-ctxt-factory (:ssl-context-factory
+                                                   @(:state webserver-context))]
                          (HttpClient. ssl-ctxt-factory)
                          (HttpClient.)))]
-          (if request-buffer-size
-            (.setRequestBufferSize client request-buffer-size)
-            (.setRequestBufferSize client config/default-request-header-buffer-size))
+          (when request-buffer-size
+            (.setRequestBufferSize client request-buffer-size))
           client))
 
       (createHttpClient []
-        (let [client (proxy-super createHttpClient)]
+        (let [client (proxy-super createHttpClient)
+              timeout (when idle-timeout
+                        (* 1000 idle-timeout))]
           (if (:follow-redirects options)
             (.setFollowRedirects client true)
             (.setFollowRedirects client false))
+          (when timeout
+            (.setIdleTimeout client timeout))
           client))
 
       (customizeProxyRequest [proxy-req req]
         (if-let [callback-fn (:callback-fn options)]
          (callback-fn proxy-req req)))
 
+      ;; The implementation of onResponseFailure is duplicated heavily from:
+      ;; https://github.com/eclipse/jetty.project/blob/jetty-9.2.10.v20150310/jetty-proxy/src/main/java/org/eclipse/jetty/proxy/AbstractProxyServlet.java#L576-L607
+      ;; The only significant difference is that a 'failure-callback-fn', if
+      ;; defined in options, is invoked just prior to completing the async
+      ;; context for cases that the response was not already committed upstream.
       (onResponseFailure [req resp proxy-resp failure]
         (do
-          (log/debug (str (.getRequestId this req) " proxying failed: " failure))
-          (when-not (.isCommitted resp)
-            (if (instance? TimeoutException failure)
-              (.setStatus resp HttpServletResponse/SC_GATEWAY_TIMEOUT)
-              (.setStatus resp HttpServletResponse/SC_BAD_GATEWAY))
-            (when-let [failure-callback-fn (:failure-callback-fn options)]
-              (failure-callback-fn req resp proxy-resp failure))
-            (.complete (.getAttribute req "org.eclipse.jetty.proxy.ProxyServlet.asyncContext")))))
-        )))
+          (let [request-id    (.getRequestId this req)
+                async-context (.getAsyncContext req)]
+            (log/debug failure (str request-id " proxying failed"))
+            (if (.isCommitted resp)
+              (try
+                (.sendError resp -1)
+                (.complete async-context)
+                (catch IOException _
+                  (log/debug failure (str request-id
+                                          " could not close the connection"))))
+              (do
+                (.resetBuffer resp)
+                (if (instance? TimeoutException failure)
+                  (.setStatus resp HttpServletResponse/SC_GATEWAY_TIMEOUT)
+                  (.setStatus resp HttpServletResponse/SC_BAD_GATEWAY))
+                (.setHeader resp
+                            (.asString HttpHeader/CONNECTION)
+                            (.asString HttpHeaderValue/CLOSE))
+                (when-let [failure-callback-fn (:failure-callback-fn options)]
+                  (failure-callback-fn req resp proxy-resp failure))
+                (.complete async-context)))))))))
 
 (schema/defn ^:always-validate
   register-endpoint!
@@ -408,7 +534,11 @@
   []
   (let [^ContextHandlerCollection chc (ContextHandlerCollection.)]
     {:handlers chc
-     :state (atom {:endpoints {}})
+     :state (atom {:endpoints {}
+                   :mbean-container nil
+                   :overrides-read-by-webserver false
+                   :overrides nil
+                   :ssl-context-factory nil})
      :server nil}))
 
 ; TODO move out of public
@@ -428,10 +558,20 @@
     (merge options overrides)))
 
 (schema/defn ^:always-validate shutdown
-  [webserver-context :- ServerContext]
+  [{:keys [server] :as webserver-context} :- ServerContext]
+  (when-let [mbean-container (:mbean-container @(:state webserver-context))]
+    (log/debug "Cleaning up JMX MBean container")
+    (.destroy mbean-container))
   (when (started? webserver-context)
     (log/info "Shutting down web server.")
-    (.stop (:server webserver-context))))
+    (try
+      (.stop server)
+      (catch TimeoutException e
+        (log/errorf e
+                    (str "Web server failed to shut down gracefully in configured "
+                         "timeout period (%s); cancelling remaining requests.")
+                    (.getStopTimeout server))))
+    (log/info "Web server shutdown")))
 
 (schema/defn ^:always-validate
   create-webserver :- ServerContext
@@ -497,7 +637,10 @@
                                  (.setHandler maybe-logged))
                                maybe-logged)]
       (.setHandler s statistics-handler)
-      (.setStopTimeout s (or shutdown-timeout default-graceful-stop-timeout))
+      (when shutdown-timeout
+        (.setStopTimeout s shutdown-timeout))
+      (when-let [script (:post-config-script options)]
+        (config/execute-post-config-script! s script))
       (assoc webserver-context :server s))))
 
 (schema/defn ^:always-validate start-webserver! :- ServerContext
@@ -550,21 +693,29 @@
                        (.setHandler (ring-handler handler)))]
     (add-handler webserver-context ctxt-handler enable-trailing-slash-redirect?)))
 
+(schema/defn ^:always-validate
+  add-websocket-handler :- ContextHandler
+  [webserver-context :- ServerContext
+   handlers :- websockets/WebsocketHandlers
+   path :- schema/Str
+   enable-trailing-slash-redirect?]
+  (let [ctxt-handler (doto (ContextHandler. path)
+                       (.setHandler (websockets/websocket-handler handlers)))]
+    (add-handler webserver-context ctxt-handler enable-trailing-slash-redirect?)))
+
 (schema/defn ^:always-validate
   add-servlet-handler :- ContextHandler
-  ([webserver-context servlet path]
-   (add-servlet-handler webserver-context servlet path {}))
-  ([webserver-context :- ServerContext
-    servlet :- Servlet
-    path :- schema/Str
-    servlet-init-params :- {schema/Any schema/Any}
-    enable-trailing-slash-redirect?]
-   (let [holder   (doto (ServletHolder. servlet)
-                    (.setInitParameters servlet-init-params))
-         handler  (doto (ServletContextHandler. ServletContextHandler/SESSIONS)
-                    (.setContextPath path)
-                    (.addServlet holder "/*"))]
-     (add-handler webserver-context handler enable-trailing-slash-redirect?))))
+  [webserver-context :- ServerContext
+   servlet :- Servlet
+   path :- schema/Str
+   servlet-init-params :- {schema/Any schema/Any}
+   enable-trailing-slash-redirect?]
+  (let [holder (doto (ServletHolder. servlet)
+                 (.setInitParameters servlet-init-params))
+        handler (doto (ServletContextHandler. ServletContextHandler/SESSIONS)
+                  (.setContextPath path)
+                  (.addServlet holder "/*"))]
+    (add-handler webserver-context handler enable-trailing-slash-redirect?)))
 
 (schema/defn ^:always-validate
   add-war-handler :- ContextHandler
@@ -615,7 +766,7 @@
     at that endpoint. Each of these maps contains the type of the
     handler under the :type key, and may contain additional information
     as well.
-    
+
     When the value of :type is :context, the endpoint information will
     be an instance of ContextEndpoint.
 
@@ -813,6 +964,19 @@
     (register-endpoint! state endpoint-map path)
     (add-ring-handler s handler path enable-redirect)))
 
+(schema/defn ^:always-validate add-websocket-handler!
+  [context
+   handlers :- websockets/WebsocketHandlers
+   path :- schema/Str
+   options :- CommonOptions]
+  (let [server-id     (:server-id options)
+        s             (get-server-context context server-id)
+        state         (:state s)
+        endpoint-map  {:type     :websocket}
+        enable-redirect  (:redirect-if-no-trailing-slash options)]
+    (register-endpoint! state endpoint-map path)
+    (add-websocket-handler s handlers path enable-redirect)))
+
 (schema/defn ^:always-validate add-servlet-handler!
   [context servlet path options :- ServletHandlerOptions]
   (let [defaults            {:servlet-init-params {}}
